//#define DEBUG

using System;
using UnityEngine;
using UnityEngine.UI;
using UnityEngine.XR;
using UnityEngine.Events;
using System;
using System.Text;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using AssetBundles;
using System.Threading;

namespace Unique
{
    public class AutoMirrorLOD : MVRScript
    {
        public string pluginAuthor = "unique";
        public string pluginName = "AutoMirrorLOD";
        public string pluginVersion = "1.0";
        public string pluginDate = "2022-03-07";
        public string pluginDescription =
 @"
 Automatically adjusts the render resolution (LOD levels) of mirror atoms based on the distance from the current camera to reduce GPU and CPU usage.
 ";
   
        private JSONStorableFloat myTime;
        public JSONStorableFloat lod0distance;
        public JSONStorableFloat lod1distance;
        public JSONStorableFloat lod2distance;
        
        private float myClock = -1.0f;
        
        private bool invalidAtom = false;
        
        // Use this for initialization
        public override void Init()
        {
            // Usage text...
            JSONStorableString usageInfoTextJSON = new JSONStorableString("", "\nAutomatically adjusts the render resolution (LOD levels) of mirror atoms based on the distance from the current camera to reduce GPU and CPU usage.");
            UIDynamicTextField usageInfoText = CreateTextField(usageInfoTextJSON);
            usageInfoText.backgroundColor = new Color32(60, 58, 210, 255);
            usageInfoText.height = 260f;
            usageInfoText.textColor = Color.white;
            
            myTime = new JSONStorableFloat("Update Speed (sec)", 0.25f, 0.0f, 10.0f, false);
            myTime.constrained = false;
            myTime.storeType = JSONStorableParam.StoreType.Full;
            RegisterFloat(myTime);
            CreateSlider(myTime, true);
            
            myClock = myTime.val;
            
            
            JSONStorableString lodSettingsTextJSON = new JSONStorableString("", "\nLOD Distance Settings");
            UIDynamicTextField lodSettingsText = CreateTextField(lodSettingsTextJSON, true);
            lodSettingsText.backgroundColor = new Color32(210, 60, 58, 255);
            lodSettingsText.height = 20f;
            lodSettingsText.textColor = Color.white;
            
            lod0distance = new JSONStorableFloat("LOD 0 Distance (4K)", 2.0f, 0.0f, 100.0f, false);
            lod0distance.constrained = true;
            lod0distance.storeType = JSONStorableParam.StoreType.Full;
            RegisterFloat(lod0distance);
            CreateSlider(lod0distance, true);
            
            float lod2min = 0.0f;
            
            if (lod0distance.val > 0.0f)
            {
                lod2min = lod0distance.val + 1.0f;
            }
            
            lod1distance = new JSONStorableFloat("LOD 1 Distance (2K)", 10.0f, lod2min, 100.0f, false);
            lod1distance.constrained = true;
            lod1distance.storeType = JSONStorableParam.StoreType.Full;
            RegisterFloat(lod1distance);
            CreateSlider(lod1distance, true);
            
            if (lod1distance.val < lod0distance.val)
            {
                lod1distance.val = lod0distance.val + 1.0f;
            }
            
            lod2distance = new JSONStorableFloat("LOD 2 Distance (1K)", 40.0f, lod1distance.val + 1.0f, 100.0f, false);
            lod2distance.constrained = true;
            lod2distance.storeType = JSONStorableParam.StoreType.Full;
            RegisterFloat(lod2distance);
            CreateSlider(lod2distance, true);
            
            if (lod2distance.val < lod1distance.val)
            {
                lod2distance.val = lod1distance.val + 1.0f;
            }
            
            JSONStorableString lodInfoTextJSON = new JSONStorableString("", "\n0.5K resolution used when further.\n\nSet LOD 0 to 0 to disable 4K.\nSet LOD 0 & 1 to 0 to disable 4K & 2K.");
            UIDynamicTextField lodInfoText = CreateTextField(lodInfoTextJSON, true);
            lodInfoText.height = 180f;
            lodInfoText.backgroundColor = new Color32(60, 140, 58, 255);
            lodInfoText.textColor = Color.white;
            
            
            if (containingAtom.type != "Glass" && containingAtom.type != "Glass-Stained" && containingAtom.type != "ReflectiveSlate" && containingAtom.type != "ReflectiveWoodPanel")
            {
                SuperController.LogMessage("AutoMirrorLOD: Atom \"" + containingAtom.type + "\" is of the wrong type. Atom type must be \"Glass\", \"Glass-Stained\", \"ReflectiveSlate\", or \"ReflectiveWoodPanel\".");
                invalidAtom = true;
            }
            else
            {
                invalidAtom = false;
            }
        }
        
        
        private string currentLodLevel = "512";
        
        void Update()
        {
            if (invalidAtom)
            {// Just skip...
                return;
            }
            
            myClock -= Time.deltaTime;
            
            if (myClock > 0.0f)
            {// wait
                return;
            }
            
            myClock += myTime.val;
            myClock = Mathf.Max(myClock, 0.0f);
            
            float lod2min = 0.0f;
            
            if (lod0distance.val > 0.0f)
            {
                lod2min = lod0distance.val + 1.0f;
            }
            
            lod1distance.min = lod2min;
            
            try
            {
                Transform cameraT = SuperController.singleton.lookCamera.transform;
                float dist = Vector3.Distance(containingAtom.transform.position, cameraT.position);
                
                var colliders = containingAtom.GetComponentsInChildren<Collider>();
                
                if (colliders.Length > 0)
                {// We have a collider, so use more accurate distance (the closest point on the mirror to eyes)...
                    dist = Vector3.Distance(colliders[0].ClosestPoint(cameraT.position), cameraT.position);
                    
                    foreach (Collider c in colliders)
                    {
                        dist = Mathf.Min(dist, Vector3.Distance(c.ClosestPoint(cameraT.position), cameraT.position));
                    }
                }
                
                
                // Collider...
                var atomControl = containingAtom.GetStorableByID("MirrorRender");
                
                if (atomControl != null)
                {
                    string newLodLevel = "512";
                    
                    if (dist < lod0distance.val)
                    {
                        newLodLevel = "4096";
                    }
                    else if (dist < lod1distance.val)
                    {
                        newLodLevel = "2048";
                    }
                    else if (dist < lod2distance.val)
                    {
                        newLodLevel = "1024";
                    }
                    else
                    {
                        newLodLevel = "512";
                    }
                    
#if DEBUG
                    if (currentLodLevel != newLodLevel)
                    {
                        SuperController.LogMessage("AutoMirrorLOD: Mirror lod level changed to  " + newLodLevel + " (distance " + dist + ")");
                    }
#endif //DEBUG

                    currentLodLevel = newLodLevel;
                    
                    atomControl.SetStringChooserParamValue("textureSize", currentLodLevel);
                }
            }
            catch (Exception e)
            {
                SuperController.LogError("Exception caught: " + e);
            }
        }
    }
}
